import functools

from toolz import merge

from series.db import Database, Column, Integer, Boolean, String, Float


@functools.total_ordering
@Database.many_to_one('Season', backref_name='episodes')
@Database.many_to_one('Series', backref_name='episodes')
class Episode(object, metaclass=Database.DefaultMeta):
    number = Column(Integer, nullable=False)
    title = Column(String)
    new = Column(Boolean, default=True)
    removed = Column(Boolean, default=False)
    subfps = Column(String)
    subdelay = Column(Float)
    metadata_fetched = Column(Boolean, default=False)
    metadata_failures = Column(Integer, default=0)
    overview = Column(String)

    def __str__(self):
        return '{} {}x{}'.format(self.series, self.season.number, self.number)

    def __repr__(self):
        return '<Episode {}x{} ({})>'.format(self.season.number, self.number,
                                             self.series.formatted_name)

    def __lt__(self, other):
        if isinstance(other, Episode):
            return (self.series < other.series or
                    self.series == other.series and
                    (self.season < other.season or
                     (self.season == other.season and
                      self.number < other.number)))

    def __eq__(self, other):
        if isinstance(other, Episode):
            return (self.series == other.series and
                    self.season == other.season and
                    self.number == other.number)

    def __hash__(self):
        return hash((self.series, self.season, self.number))

    @property
    def info(self):
        return dict(
            id=self.id,
            series=self.series.formatted_name,
            season=self.season.number,
            episode=self.number,
            title=self.title,
            new=self.new,
            last_watched=self.last_watched_string,
            subfps=self.subfps,
            subdelay=self.subdelay,
            overview=self.overview,
            type='episode',
        )

    @property
    def ext_info(self):
        ext = dict(season=self.season.ext_info)
        return merge(self.info, ext)

    @property
    def last_watched(self):
        if self.watch_history.count() > 0:
            return sorted(self.watch_history)[-1]

    @property
    def last_watched_string(self):
        last_watched = self.last_watched
        return (last_watched.begin.strftime('%d.%m.%y %H:%M') if last_watched
                else 'Never')

    @property
    def subfps_fallback(self):
        return self.subfps or self.season.subfps

    @property
    def resume_position(self):
        if self.last_watched:
            return self.last_watched.stopped_at

__all__ = ['Episode']



"""add ReleaseMonitor.last_torrent_search_stamp

Revision ID: 2ee6e71eb2f
Revises: 36bd688b04b
Create Date: 2014-12-16 18:31:20.327836

"""

# revision identifiers, used by Alembic.
revision = '2ee6e71eb2f'
down_revision = '36bd688b04b'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release_monitor', sa.Column('last_torrent_search_stamp', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('release_monitor', 'last_torrent_search_stamp')
    ### end Alembic commands ###



from golgi import configurable

from series.library.db import FileDatabase
from series.library.library_facade import LibraryFacade
from series.library.player import Player
from series.app import App


@configurable(library=['db_path', 'run', 'omit'])
class SeriesLibraryD(App):
    _components = ['rest_api', 'metadata']

    def __init__(self):
        self.db = FileDatabase(self._db_path)
        self.library = LibraryFacade(self.db)
        self.player = Player(self.library)
        super().__init__('library', self._run, self._omit,
                         c_args=(self.library, self.player), name='SeriesLibD')

    def _cleanup(self):
        self.db.session.commit()
        self.db.disconnect()

__all__ = ['SeriesLibraryD']



from golgi.config.options import BoolConfigOption

metadata = dict(parents=['series.store_episode', 'series.subsync',
                         'tek_utils.extract'])


def reset_config():
    archive_ext = '(rar|zip|7z|tar\.[bg]z2?)'
    episode_ext = '(mkv|avi|srt)'
    return {
        'handle_episode': dict(
            pretend=False, archive_ext=archive_ext, episode_ext=episode_ext,
            subtitles=BoolConfigOption(True, no=True)
        )
    }



import re
import random

import requests

from series.logging import Logging
from tek.tools import find, resolve_redirect, repr_params, camelcaseify
from tek.util.decorator import generated_list
from golgi.config import configurable

from tek_utils.sharehoster import accepted_domains
from tek_utils.sharehoster.torrent import is_torrent

from series.get.model.release import ReleaseFactory
from series.util import domain


def _valid_entry(entry):
    return ('content' in entry and entry['content'] and
            'value' in entry['content'][0])


class FeedEntry(Logging):

    def __init__(self, release, links=None, is_nuke=False):
        self.release = release
        self.links = links or []
        self.is_nuke = is_nuke

    def __repr__(self):
        params = repr_params(self.release, self.links, self.is_nuke)
        return '{}{}'.format(self.__class__.__name__, params)

    def __str__(self):
        return '{}: {}'.format(self.__class__.__name__, str(self.release))

    def resolve_redirects(self):
        for i, link in enumerate(self.links):
            if not is_torrent(link):
                try:
                    self.links[i] = resolve_redirect(link)
                except requests.RequestException as e:
                    self.log.error('FeedEntry: {}'.format(e))

    @property
    def is_fix(self):
        return self.release.is_fix


class LinkFinder(object):
    fallback_template = ('{{}} {provider}[^<]+?<a[^>]*?href="(?P<link>[^"]+)"')
    re_template = ''

    def __init__(self, provider):
        self._provider = provider

    def process(self, entry, title):
        if _valid_entry(entry):
            return (self.source_specific(entry, title) or
                    self.fallback(entry, title))

    def source_specific(self, entry, title):
        if self.re_template:
            content = entry['content'][0]['value']
            m = re.search(self.regex(title), content, re.DOTALL | re.I)
            if m:
                return m.group('link')

    def fallback(self, entry, title):
        m = re.search(self.fallback_regex(title),
                      entry['summary_detail']['value'], re.DOTALL | re.I)
        if m:
            return m.group('link')

    def regex(self, title):
        return self.re_template.format(title=title, provider=self._provider)

    def fallback_regex(self, title):
        return self.fallback_template.format(title=title,
                                             provider=self._provider)


class WrzkoLinkFinder(LinkFinder):
    re_template = ('<strong>{title}.+?{provider}.+?<a[^>]*?href=' +
                   '"(?P<link>[^"]+)"')


class RlsbbLinkFinder(LinkFinder):
    re_template = '<strong>{title}.+?<a[^>]*?href="(?P<link>[^"]+)">{provider}'


class FeedTorrentFinder(object):

    def process(self, entry, title):
        link = entry.get('magneturi', entry.get('link'))
        if is_torrent(link):
            return link


class EntryParserMeta(type):

    def __init__(cls, name, bases, dct):
        date_enum = cls.re_separator.join([r'\d{4}', r'\d{2}', r'\d{2}'])
        parts = (cls.re_parts_pre + [cls.enum.format(date=date_enum)] +
                 cls.re_parts_post)
        rex = cls.re_separator.join(parts)
        cls.release_re_raw = r'{}{}'.format(rex, cls.group)
        cls.release_re = re.compile('^{}$'.format(cls.release_re_raw), re.I)
        super(EntryParserMeta, cls).__init__(name, bases, dct)


class EntryParser(object, metaclass=EntryParserMeta):
    _name = r'(?P<name>.+?)'
    enum = r'(?P<enum>S?\d{{1,2}}?[xE]\d{{1,2}}?|{date})((-|E)\d{{1,2}}?)?'
    _flags = r'(?P<flags>.+?)?'
    group = r'(-(?P<group>[^-]+?))?'
    re_parts_pre = [_name]
    re_parts_post = [_flags]
    re_separator = '\.'
    domain = ''
    _release_fact = ReleaseFactory()

    def __init__(self, entry, link_finders):
        self.entry = entry
        self._link_finders = link_finders
        self._releases = None

    @property
    def valid(self):
        return self.release is not None

    @property
    def release(self):
        return find(lambda r: r.hd_series, self.releases)

    @property
    def title(self):
        return self.release.title if self.release else ''

    @property
    def titles(self):
        return [self.entry['title']]

    @property
    def releases(self):
        if self._releases is None:
            self._releases = self._find_releases()
        return self._releases

    def _find_releases(self):
        return self._releases_from_titles()

    @property
    def links(self):
        return [_f for _f in [self._find_link(finder) for finder in
                              self._link_finders] if _f]

    def _find_link(self, finder):
        return finder.process(self.entry, self.title)

    @generated_list
    def _releases_from_titles(self):
        for title in self.titles:
            release = self._release_from_title(title)
            if release:
                yield release

    def _release_from_title(self, title):
        title = title.lower().strip()
        match = self.release_re.match(title)
        if match:
            return self._release_fact.from_title_match(title, match)
        else:
            text = 'Couldn\'t match title "{}"'.format(title)
            self.log.error(text)


class WrzkoParser(EntryParser):
    domain = 'wrzko'

    @property
    def titles(self):
        return [t.lower().strip() for t in self.entry['title'].split('&')]


class RlsbbParser(EntryParser):
    domain = 'rlsbb'

    @property
    def titles(self):
        title_re = '<strong> *({}) *<'.format(self.release_re_raw)
        if _valid_entry(self.entry):
            content = self.entry['content'][0].value
            candidates = re.findall(title_re, content)
            return [c[0].lower().strip() for c in candidates if c]
        else:
            return []


class EzrssParser(EntryParser):
    domain = 'ezrss'

    @property
    def titles(self):
        title = self.entry.get('filename', '').rsplit('.', 2)[0]
        return [title]


class ShowrssParser(EntryParser):
    domain = 'showrss'

    @property
    def titles(self):
        title = self.entry.get('showrss_rawtitle', '').replace(' ', '.')
        title = '-'.join(title.rsplit('.', 1))
        return [title]


@configurable(general=['verbose'])
class FeedEntryFactory(Logging):

    def __init__(self, rss_url):
        self.domain = domain(rss_url)
        self._Parser = globals().get(
            camelcaseify('{}_parser'.format(self.domain)), EntryParser
        )
        _LinkFinder = globals().get(
            camelcaseify('{}_link_finder'.format(self.domain)), LinkFinder)
        self._link_finders = [_LinkFinder(p) for p in accepted_domains()]
        self._link_finders += [FeedTorrentFinder()]

    def process_items(self, items, want_hd=True):
        entries = [self.process_item(item, want_hd) for item in items]
        return [e for e in entries if e is not None]

    def process_item(self, item, want_hd=True):
        parser = self._Parser(item, self._link_finders)
        if parser.valid:
            return FeedEntry(parser.release, links=parser.links)
        else:
            if self._verbose:
                msg = 'Discarding feed entry "{}" with releases:'
                self.log.warning(msg.format(item['title']))
                for release in parser.releases:
                    self.log.warning(release)

    def from_title(self, title, want_hd=True):
        return self.process_item(dict(title=title), want_hd)


class FeedEntryHasher(object):

    def __init__(self, rss_url):
        pass

    def __call__(self, entry):
        if _valid_entry(entry):
            return hash(tuple(entry['content'][0]['value']))
        else:
            return random.randint(0, 1000000)



"""add search_name to Release and Show

Revision ID: 218daaf29e40
Revises: 18b0a0874c5a
Create Date: 2016-11-15 20:00:04.915748

"""

# revision identifiers, used by Alembic.
revision = '218daaf29e40'
down_revision = '18b0a0874c5a'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('release', sa.Column('search_name', sa.String(), nullable=True))
    op.add_column('show', sa.Column('search_name', sa.String(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    pass
    ### end Alembic commands ###



import operator
from datetime import datetime, timedelta

from series.get.handler import ShowHandler
from series.get.tvdb import Tvdb

from series.condition import HandlerCondition, SimpleCondition, LambdaCondition
from series.util import datetime_to_unix


class AirsToday(HandlerCondition):

    @property
    def _today_thresh(self):
        return datetime.now() - timedelta(hours=16)

    def _check(self, show):
        return show.next_episode_date < self._today_thresh

    def ev(self, show):
        return show.has_next_episode and self._check(show)

    def describe(self, show, target):
        match = self.ev(show)
        good = match == target
        today = (show.next_episode_date > self._today_thresh) and match
        desc = ('today' if today else
                show.next_episode_date.strftime('%F') if
                show.next_episode_stamp > 0 else
                'no date')
        return 'airs today[{}]'.format(self._paint(desc, good))


class CanCatchUp(SimpleCondition):
    ''' *current* is the last aired episode, which includes those aired
    within 24 hours in the future.
    *latest* is the newest release in the db.
    '''

    def __init__(self, latest) -> None:
        self._latest = latest

    def latest_episode(self, show):
        return show.latest_episode_m

    def latest_release(self, show):
        return self._latest(show)

    def ev(self, show):
        return (self.latest_episode(show) & self.latest_release(show)
                ).map2(operator.gt).true

    @property
    def _desc(self):
        return 'latest > downloaded'

    def _repr(self, show, match):
        op = '>' if match else '<'
        def has(release):
            return '{} {} {}'.format(self.latest_episode(show) | -1, op,
                                     release)
        return self.latest_release(show) / has | 'no latest episode'


class ShowScheduler(ShowHandler, Tvdb):

    def __init__(self, releases, shows, **kw):
        super().__init__(shows, 60, 'show scheduler', cooldown=3600, **kw)
        self._releases = releases

    def _handle(self, show):
        airdate = self.tvdb.airdate(show, show.season, show.next_episode)
        if (show.has_next_episode and airdate and airdate !=
                show.next_episode_date):
            self._handle_invalid_show(show, airdate)
        else:
            self._handle_valid_show(show)

    def _handle_valid_show(self, show):
        def catch_up(latest_release, latest_episode):
            for episode in range(latest_release + 1, latest_episode + 1):
                self._schedule(show, show.latest_season, episode)
        def schedule_next():
            self._schedule(show, show.season, show.next_episode)
        return (
            schedule_next()
            if AirsToday().ev(show) else
            (self._latest(show) & show.latest_episode_m).map2(catch_up)
        )

    def _handle_invalid_show(self, show, airdate):
        msg = '{} {}x{} was rescheduled: {} => {}'
        self.log.error(msg.format(show.name, show.season, show.next_episode,
                                  show.next_episode_day, airdate.date()))
        self._update(show, next_episode_stamp=datetime_to_unix(airdate))

    @property
    def _conditions(self):
        return (LambdaCondition('next has no release', self._no_next_release) &
                (AirsToday() | CanCatchUp(self._latest)))

    def _latest(self, show):
        return self._releases.latest_for_season(show.canonical_name,
                                                show.latest_season)

    def _no_next_release(self, show):
        return not self._latest(show).contains(show.next_episode)

    def _schedule(self, show, season, episode):
        airdate = self.tvdb.airdate(show, season, episode)
        msg = 'Scheduling release "{} {}x{}" on {}'.format(
            show.name,
            season,
            episode,
            airdate
        )
        self.log.info(msg)
        self._releases.create(show.canonical_name, season, episode, airdate,
                              downgrade_after=show.downgrade_after,
                              search_name=show.search_name)
        if show.season > show.latest_season:
            self._shows.update_by_id(show.id, latest_season=show.season)
        self._commit()

__all__ = ['ShowScheduler']



from series.get.tvrage_facade import TVRageFacade
from series.etvdb import ETVDBFacade
from series.logging import Logging

from golgi.config import configurable, ConfigError


@configurable(show_planner=['show_db'])
class Tvdb(Logging):
    _tvdbs = dict(etvdb=ETVDBFacade, tvrage=TVRageFacade)
    _tvdb = None

    @property
    def tvdb(self):
        if self._tvdb is None:
            if self._show_db not in self._tvdbs:
                msg = 'Invalid config value for show_db: {} (allowed: {})'
                raise ConfigError(msg.format(', '.join(self._tvdbs.keys())))
            else:
                self._tvdb = self._tvdbs[self._show_db]()
        return self._tvdb

    @property
    def use_etvdb(self):
        return self._show_db == 'etvdb'

__all__ = ['Tvdb']
